#include <iostream>
#include <string>
#include <vector>
#include <cstdint>
#include <cstring>
#include <cmath>
#include <malloc.h>
#include <iomanip>
#include <gccore.h>
#include <algorithm>
#include <wiiuse/wpad.h>

#define DEFAULT_FIFO_SIZE	(256*1024)

static GXRModeObj *rmode = NULL;
static void *frameBuffer[2] = { NULL, NULL};

u8 texarray[0x400] ATTRIBUTE_ALIGN(32);

f32 cubevtxarray[] ATTRIBUTE_ALIGN(32) = {
	-1.0f, -1.0f, -1.0f,
	-1.0f, -1.0f,  1.0f,
	-1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f,  1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f,  1.0f,
	 1.0f,  1.0f, -1.0f,
	 1.0f,  1.0f,  1.0f
};

f32 cubetexarray[32] ATTRIBUTE_ALIGN(32) = {
	0.0f, 0.0f,
	1.0f, 0.0f,
	1.0f, 1.0f,
	0.0f, 1.0f
};

//---------------------------------------------------------------------------------
int main(int argc, char **argv) {
//---------------------------------------------------------------------------------
	
	std::fill_n(texarray, 0x400, 0xff);
	
	f32 yscale;
	u32 xfbHeight;
	u32 fb = 0;
	f32 rquad = 0.0f;
	u32 first_frame = 1;
	GXTexObj texture;
	Mtx view; // view and perspective matrices
	Mtx model, modelview;
	Mtx44 perspective;
	void *gpfifo = NULL;
	GXColor background = {50, 50, 50, 0xff};
	guVector cam  = {0.0F, 0.0F, -12.0F},
			 up   = {0.0F, 1.0F,   0.0F},
			 look = {0.0F, 0.0F,  -1.0F};

	
	// Initialise the video system
	VIDEO_Init();

	// Initialise the attached controllers
	PAD_Init();

	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);

	// allocate the fifo buffer
	gpfifo = memalign(32, DEFAULT_FIFO_SIZE);
	memset(gpfifo, 0, DEFAULT_FIFO_SIZE);

	// allocate 2 framebuffers for double buffering
	frameBuffer[0] = SYS_AllocateFramebuffer(rmode);
	frameBuffer[1] = SYS_AllocateFramebuffer(rmode);

	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);

	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(frameBuffer[fb]);
	// Flush the video register changes to the hardware
	VIDEO_Flush();
	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(rmode->viTVMode & VI_NON_INTERLACE) VIDEO_WaitVSync();

	fb ^= 1;

	// init the flipper
	GX_Init(gpfifo,DEFAULT_FIFO_SIZE);

	// clears the bg to color and clears the z buffer
	GX_SetCopyClear(background, 0x00ffffff);

	// other gx setup
	GX_SetViewport(0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1);
	yscale = GX_GetYScaleFactor(rmode->efbHeight, rmode->xfbHeight);
	xfbHeight = GX_SetDispCopyYScale(yscale);
	GX_SetScissor(0, 0, rmode->fbWidth, rmode->efbHeight);
	GX_SetDispCopySrc(0, 0, rmode->fbWidth, rmode->efbHeight);
	GX_SetDispCopyDst(rmode->fbWidth, xfbHeight);
	GX_SetCopyFilter(rmode->aa, rmode->sample_pattern, GX_TRUE, rmode->vfilter);
	GX_SetFieldMode(rmode->field_rendering, ((rmode->viHeight == (2 * rmode->xfbHeight))? GX_ENABLE : GX_DISABLE));

	if (rmode->aa) {
		GX_SetPixelFmt(GX_PF_RGB565_Z16, GX_ZC_LINEAR);
	} else {
		GX_SetPixelFmt(GX_PF_RGB8_Z24, GX_ZC_LINEAR);
	}

	GX_SetCullMode(GX_CULL_BACK);
	GX_CopyDisp(frameBuffer[fb], GX_TRUE);
	GX_SetDispCopyGamma(GX_GM_1_0);

	// setup the vertex attribute table
	GX_ClearVtxDesc();

	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XYZ, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
	
	GX_SetVtxDesc(GX_VA_POS, GX_INDEX8);
	GX_SetVtxDesc(GX_VA_TEX0, GX_INDEX8);
	
	GX_SetArray(GX_VA_POS, cubevtxarray, 3 * sizeof(f32));
	GX_SetArray(GX_VA_TEX0, cubetexarray, 2 * sizeof(f32));

	// set number of rasterized color channels
	GX_SetNumChans(1);

	//set number of textures to generate
	GX_SetNumTexGens(1);

	// setup texture coordinate generation
	// args: texcoord slot 0-7, matrix type, source to generate texture coordinates from, matrix to use
	GX_SetTexCoordGen(GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);

	GX_InvVtxCache();
	GX_InvalidateTexAll();
	
	//load image data into the texture slot
	GX_InitTexObj(&texture, texarray, 16, 16, GX_TF_RGBA8, GX_REPEAT, GX_REPEAT, GX_TRUE);
	GX_InitTexObjFilterMode(&texture, GX_NEAR, GX_NEAR);
	
	// setup our camera at the origin
	// looking down the -z axis with y up
	guLookAt(view, &cam, &up, &look);

	// setup our projection matrix
	// this creates a perspective matrix with a view angle of 90,
	// and aspect ratio based on the display resolution
	float w = rmode->viWidth;
	float h = rmode->viHeight;
	guPerspective(perspective, 45, w / h, 0.1F, 300.0F);
	GX_LoadProjectionMtx(perspective, GX_PERSPECTIVE);

	guVector cubeAxis = {0, 1, 0};

	int16_t x{};

	while(1) {

		PAD_ScanPads();
		uint16_t buttons = PAD_ButtonsDown(0);
		if (buttons & PAD_BUTTON_START) break;
		
		if (buttons & (PAD_BUTTON_LEFT | PAD_BUTTON_RIGHT)) {
			if (buttons & PAD_BUTTON_RIGHT) {
				++x;
			} else {
				--x;
			}
			if (x < 0) {
				x = 0x400 - 1;
			}
			if (x >= 0x400) {
				x = 0;
			}
			texarray[x] = 0;
			GX_InvalidateTexAll();
		}

		GX_LoadTexObj(&texture, GX_TEXMAP0);

		float offset = sin(rquad / 2 + 0.15) * 4;
		look = {0.0F, offset, -1.0F};
		guLookAt(view, &cam, &up, &look);
		offset = sin(rquad / 2) * 4;
		
		guMtxIdentity(model);
		guMtxRotAxisRad(model, &cubeAxis, rquad);
		guMtxTransApply(model, model, 0.0f, offset, 0.0f);
		guMtxConcat(view, model, modelview);
		// load the modelview matrix into matrix memory
		GX_LoadPosMtxImm(modelview, GX_PNMTX0);

		GX_Begin(GX_QUADS, GX_VTXFMT0, 24);			// Draw a Cube
			//top
			GX_Position1x8(7);
			GX_TexCoord1x8(0);
			GX_Position1x8(3);
			GX_TexCoord1x8(1);
			GX_Position1x8(2);
			GX_TexCoord1x8(2);
			GX_Position1x8(6);
			GX_TexCoord1x8(3);
			//bottom
			GX_Position1x8(5);
			GX_TexCoord1x8(0);
			GX_Position1x8(4);
			GX_TexCoord1x8(1);
			GX_Position1x8(0);
			GX_Position1x8(2);
			GX_Position1x8(1);
			GX_TexCoord1x8(3);
			//front
			GX_Position1x8(6);
			GX_TexCoord1x8(0);
			GX_Position1x8(2);
			GX_TexCoord1x8(1);
			GX_Position1x8(0);
			GX_Position1x8(2);
			GX_Position1x8(4);
			GX_TexCoord1x8(3);
			//back
			GX_Position1x8(3);
			GX_TexCoord1x8(0);
			GX_Position1x8(7);
			GX_TexCoord1x8(1);
			GX_Position1x8(5);
			GX_Position1x8(2);
			GX_Position1x8(1);
			GX_TexCoord1x8(3);
			//right
			GX_Position1x8(2);
			GX_TexCoord1x8(0);
			GX_Position1x8(3);
			GX_TexCoord1x8(1);
			GX_Position1x8(1);
			GX_Position1x8(2);
			GX_Position1x8(0);
			GX_TexCoord1x8(3);
			//left
			GX_Position1x8(7);
			GX_TexCoord1x8(0);
			GX_Position1x8(6);
			GX_TexCoord1x8(1);
			GX_Position1x8(4);
			GX_Position1x8(2);
			GX_Position1x8(5);
			GX_TexCoord1x8(3);
		GX_End(); // Done Drawing The Cube

		GX_SetZMode(GX_TRUE, GX_LEQUAL, GX_TRUE);
		GX_CopyDisp(frameBuffer[fb],GX_TRUE);

		GX_DrawDone();

		VIDEO_SetNextFramebuffer(frameBuffer[fb]);
		if(first_frame) {
			first_frame = 0;
			VIDEO_SetBlack(FALSE);
		}
		VIDEO_Flush();
		VIDEO_WaitVSync();
		fb ^= 1;

		rquad -= 0.034f; // Decrease The Rotation Variable For The Cube
	}
	
	return 0;
}